
; ************************************************* Macros for AVR assembler programming ********************************************

; ********************************************************** Alternative basic macros ************************************************************

; *************************************************** Release 4.1, May 31, 2019 *****************************************************

; These macros accept 16-bit registers (word registers) as parameters

; All 16-bit register pairs in the AVR general-purpose register file are to be designated as R0W, R2W, etc. to R30W

; The 16-bit registers of the virtual machine as defined by the basic macros are to be designated 

; as AW, BW, CW, XW, YW, ZW

; Ths is due to the idiosyncrasies of the AVR assembler that does not accept 

; build-in keywords and pre-defined mnemonics as user-defined symbols

; The macros defined here are thought of as examples, too, demonstrating

; how macros with 16-bit register parameters are to be declared  

; CAUTION: To be used in a project, the source file of the basic macros (e.g., AVR_basic_macros_4.1)

; must be included, too


; ******* The 16-bit (word) registers declared: 

.equ r0w 	= 0
.equ r2w 	= 2
.equ r4w 	= 4
.equ r6w 	= 6
.equ r8w 	= 8
.equ r10w 	= 10
.equ r12w 	= 12
.equ r14w 	= 14
.equ r16w 	= 16
.equ r18w 	= 18
.equ r20w 	= 20
.equ r22w 	= 22
.equ r24w 	= 24
.equ r26w 	= 26
.equ r28w 	= 28
.equ r30w 	= 30


.equ aw 	= r24w
.equ bw 	= r22w
.equ cw 	= r20w
.equ xw 	= r26w
.equ yw 	= r28w
.equ zw 	= r30w


; **********************************************************************************************


; 				**************************** Basic transports ****************************

;						All 16-bit registers r0w to r30w, AW, BW, CW, XW, YW, ZW

.macro gstw			; Store the contents of a 16-bit register as a 16-bit word at a general address		
.if @1   ==  r0w  	; general_adrs,16_bit_reg
gwst @0,r1,r0

.elif @1 ==  r2w
gwst @0,r3,r2

.elif @1 ==  r4w
gwst @0,r5,r4

.elif @1 ==  r6w
gwst @0,r7,r6

.elif @1 ==  r8w
gwst @0,r9,r8

.elif @1 ==  r10w
gwst @0,r11,r10

.elif @1 ==  r12w
gwst @0,r13,r12

.elif @1 ==  r14w
gwst @0,r15,r14

.elif @1 ==  r16w
gwst @0,r17,r16

.elif @1 ==  r18w
gwst @0,r19,r18

.elif @1 ==  r20w
gwst @0,r21,r20

.elif @1 ==  r22w
gwst @0,r23,r22

.elif @1 ==  r24w
gwst @0,r25,r24

.elif @1 ==  r26w
gwst @0,r27,r26

.elif @1 ==  r28w
gwst @0,r29,r28

.elif @1 ==  r30w
gwst @0,r31,r30

.endif
.endmacro


.macro gldw			; Load the 16-bit register with the 16-bit word at a general address		
.if @0   ==  r0w  	; 16-bit_reg,general_adrs
gwld r1,r0,@1

.elif @0 ==  r2w
gwld r3,r2,@1

.elif @0 ==  r4w
gwld r5,r4,@1

.elif @0 ==  r6w
gwld r7,r6,@1

.elif @0 ==  r8w
gwld r9,r8,@1

.elif @0 ==  r10w
gwld r11,r10,@1

.elif @0 ==  r12w
gwld r13,r12,@1

.elif @0 ==  r14w
gwld r15,r14,@1

.elif @0 ==  r16w
gwld r17,r16,@1

.elif @0 ==  r18w
gwld r19,r18,@1

.elif @0 ==  r20w
gwld r21,r20,@1

.elif @0 ==  r22w
gwld r23,r22,@1

.elif @0 ==  r24w
gwld r25,r24,@1

.elif @0 ==  r26w
gwld r27,r26,@1

.elif @0 ==  r28w
gwld r29,r28,@1

.elif @0 ==  r30w
gwld r31,r30,@1

.endif
.endmacro




; **********************************************************************************************

; 					Access 16-bit words in SRAM by indirect addressing (registers AW, BW, CW, XW, YW, ZW)


.macro ldiw			; Load indirect. <16_bit_reg> := <<base_adrs>>		
.if @0   ==  aw  		
ldiwa @1

.elif @0 ==  bw
ldiwb @1

.elif @0 ==  cw
ldiwc @1

.elif @0 ==  xw
ldiwx @1

.elif @0 ==  yw
ldiwy @1

.elif @0 ==  zw
ldiwz @1

.endif
.endmacro


.macro ldiwinc			; Load indirect. Post-increment the address pointer 		
.if @0   ==  aw  		; <16_bit_reg> := <<base_adrs>>; <<base_adrs>> .= <<base_adrs>>+2  			V 		
ldiwinca @1

.elif @0 ==  bw
ldiwincb @1

.elif @0 ==  cw
ldiwincc @1

.elif @0 ==  xw
ldiwincx @1

.elif @0 ==  yw
ldiwincy @1

.elif @0 ==  zw
ldiwincz @1

.endif
.endmacro


.macro ldiwdec			; Pre-decremet the address pointer. Then load indirect Post-decrement the address pointer 		
.if @0   ==  aw  		; <<base_adrs>> .= <<base_adrs>>-2; <16_bit_reg> := <<base_adrs>> 		
ldiwdeca @1

.elif @0 ==  bw
ldiwdecb @1

.elif @0 ==  cw
ldiwdecc @1

.elif @0 ==  xw
ldiwdecx @1

.elif @0 ==  yw
ldiwdecy @1

.elif @0 ==  zw
ldiwdecz @1

.endif
.endmacro


.macro ldiwblit			; Load indirect. <16_bit_reg> := <<base_adrs> + displacement_literal>		
.if @0   ==  aw  		
ldiwba @1,@2

.elif @0 ==  bw
ldiwbb @1,@2

.elif @0 ==  cw
ldiwbc @1,@2

.elif @0 ==  xw
ldiwbx @1,@2

.elif @0 ==  yw
ldiwby @1,@2

.elif @0 ==  zw
ldiwbz @1,@2

.endif
.endmacro


.macro ldiwbd			; Load indirect. <16_bit_reg> := <<base_adrs> + <displacement_adrs>>		
.if @0   ==  aw  		
ldiwbda @1,@2

.elif @0 ==  bw
ldiwbdb @1,@2

.elif @0 ==  cw
ldiwbdc @1,@2

.elif @0 ==  xw
ldiwbdx @1,@2

.elif @0 ==  yw
ldiwbdy @1,@2

.elif @0 ==  zw
ldiwbdz @1,@2

.endif
.endmacro


.macro stiw			; Store indirect. <<base_adrs>> := <16_bit_reg> 		
.if @1   ==  aw  		
stiwa @0

.elif @1 ==  bw
stiwb @0

.elif @1 ==  cw
stiwc @0

.elif @1 ==  xw
stiwx @0

.elif @1 ==  yw
stiwy @0

.elif @1 ==  zw
stiwz @0

.endif
.endmacro


.macro stiwinc			; Store indirect. Post-increment the address pointer 		
.if @1   ==  aw  		; <<base_adrs>> := <16_bit_reg>; <<base_adrs>> .= <<base_adrs>>+2  			V 		
stiwinca @0

.elif @1 ==  bw
stiwincb @0

.elif @1 ==  cw
stiwincc @0

.elif @1 ==  xw
stiwincx @0

.elif @1 ==  yw
stiwincy @0

.elif @1 ==  zw
stiwincz @0

.endif
.endmacro


.macro stiwdec			; Pre-decremet the address pointer. Then store indirect Post-decrement the address pointer 		
.if @1   ==  aw  		; <<base_adrs>> .= <<base_adrs>>-2; <<base_adrs>> := <16_bit_reg> 		
stiwdeca @0

.elif @1 ==  bw
stiwdecb @0

.elif @1 ==  cw
stiwdecc @0

.elif @1 ==  xw
stiwdecx @0

.elif @1 ==  yw
stiwdecy @0

.elif @1 ==  zw
stiwdecz @0

.endif
.endmacro


.macro stiwblit			; Store indirect. <<base_adrs> + displacement_literal>:= <16_bit_reg>		
.if @2   ==  aw  		
stiwba @0,@1

.elif @2 ==  bw
stiwbb @0,@1

.elif @2 ==  cw
stiwbc @0,@1

.elif @2 ==  xw
stiwbx @0,@1

.elif @2 ==  yw
stiwby @0,@1

.elif @2 ==  zw
stiwbz @0,@1

.endif
.endmacro


.macro stiwbd			; Store indirect. <<base_adrs> + <displacement_adrs> := <16_bit_reg>		
.if @2   ==  aw  		
stiwbda @0,@1

.elif @2 ==  bw
stiwbdb @0,@1

.elif @2 ==  cw
stiwbdc @0,@1

.elif @2 ==  xw
stiwbdx @0,@1

.elif @2 ==  yw
stiwbdy @0,@1

.elif @2 ==  zw
stiwbdz @0,@1

.endif
.endmacro



; **********************************************************************************************

; 					Clear a 16-bit register. All 16-bit registers r0w to r30w, AW to ZW


.macro clearw			; Clear the 16_bit_reg
.if @0   ==  r0w  		 
eor r1,r1
eor r0,r0

.elif @0 ==  r2w
eor r3,r3
eor r2,r2

.elif @0 ==  r4w
eor r5,r5
eor r4,r4

.elif @0 ==  r6w
eor r7,r7
eor r6,r6

.elif @0 ==  r8w
eor r9,r9
eor r8,r8

.elif @0 ==  r10w
eor r11,r11
eor r10,r10

.elif @0 ==  r12w
eor r13,r13
eor r12,r12

.elif @0 ==  r14w
eor r15,r15
eor r14,r14

.elif @0 ==  r16w
eor r17,r17
eor r16,r16

.elif @0 ==  r18w
eor r19,r19
eor r18,r18

.elif @0 ==  r20w
eor r21,r21
eor r20,r20

.elif @0 ==  r22w
eor r23,r23
eor r22,r22

.elif @0 ==  r24w
eor r25,r25
eor r24,r24

.elif @0 ==  r26w
eor r27,r27
eor r26,r26

.elif @0 ==  r28w
eor r29,r29
eor r28,r28

.elif @0 ==  r30w
eor r31,r31
eor r30,r30

.endif
.endmacro



; **********************************************************************************************

; 					Load or store a 16-bit word (registers AW, BW, CW, XW, YW, ZW)



.macro ldw			; Load. <16_bit_reg> := <general_mem_adrs>		
.if @0   ==  aw  		
lda @1

.elif @0 ==  bw
ldb @1

.elif @0 ==  cw
ldc @1

.elif @0 ==  xw
ldx @1

.elif @0 ==  yw
ldy @1

.elif @0 ==  zw
ldz @1

.endif
.endmacro


.macro ldbyw		; Load with a zero-extended byte out of the SRAM <		
.if @0   ==  aw  	; registers AW, BW, CW		
ldbya @1

.elif @0 ==  bw
ldbyb @1

.elif @0 ==  cw
ldbyc @1

.endif
.endmacro


.macro ldxw			; Load. General memory address in XW		
.if @0   ==  aw  	; registers AW, BW, CW, YW, ZW		
ldxa

.elif @0 ==  bw
ldxb

.elif @0 ==  cw
ldxc

.elif @0 ==  yw
ldxy

.elif @0 ==  zw
ldxz

.endif
.endmacro


.macro ldyw			; Load. General memory address in YW		
.if @0   ==  aw  	; registers AW, BW, CW, XW, ZW		
ldya

.elif @0 ==  bw
ldyb

.elif @0 ==  cw
ldyc

.elif @0 ==  yw
ldyx

.elif @0 ==  zw
ldyz

.endif
.endmacro


.macro ldzw			; Load. General memory address in ZW		
.if @0   ==  aw  	; registers AW, BW, CW, XW, YW		
ldza

.elif @0 ==  bw
ldzb

.elif @0 ==  cw
ldzc

.elif @0 ==  xw
ldzx

.elif @0 ==  yw
ldzy

.endif
.endmacro


.macro stw			; Store. <general_mem_adrs> := <16_bit_reg> 
.if @0   ==  aw  		
sta @1

.elif @0 ==  bw
stb @1

.elif @0 ==  cw
stc @1

.elif @0 ==  xw
stx @1

.elif @0 ==  yw
sty @1

.elif @0 ==  zw
stz @1

.endif
.endmacro


.macro stxw			; Store. General memory address in XW		
.if @0   ==  aw  	; registers AW, BW, CW, YW, ZW		
stxa

.elif @0 ==  bw
stxb

.elif @0 ==  cw
stxc

.elif @0 ==  yw
stxy

.elif @0 ==  zw
stxz

.endif
.endmacro


.macro styw			; Store. General memory address in YW		
.if @0   ==  aw  	; registers AW, BW, CW, XW, ZW		
stya

.elif @0 ==  bw
styb

.elif @0 ==  cw
styc

.elif @0 ==  yw
styx

.elif @0 ==  zw
styz

.endif
.endmacro


.macro stzw			; Store. General memory address in ZW		
.if @0   ==  aw  	; registers AW, BW, CW, XW, YW		
stza

.elif @0 ==  bw
stzb

.elif @0 ==  cw
stzc

.elif @0 ==  xw
stzx

.elif @0 ==  yw
stzy

.endif
.endmacro


.macro ldxincw			; Load. General memory address in XW. Post-incremet XW by 2		
.if @0   ==  aw  		; registers AW, BW, CW		
ldxinca

.elif @0 ==  bw
ldxincb

.elif @0 ==  cw
ldxincc

.endif
.endmacro


.macro ldyincw			; Load. General memory address in YW. Post-incremet YW by 2		
.if @0   ==  aw  		; registers AW, BW, CW		
ldyinca

.elif @0 ==  bw
ldyincb

.elif @0 ==  cw
ldyincc

.endif
.endmacro



.macro stxincw			; Store. General memory address in XW. Post-incremet XW by 2		
.if @0   ==  aw  		; registers AW, BW, CW		
stxinca

.elif @0 ==  bw
stxincb

.elif @0 ==  cw
stxincc

.endif
.endmacro


.macro styincw			; Store. General memory address in YW. Post-incremet YW by 2		
.if @0   ==  aw 		; registers AW, BW, CW		
styinca

.elif @0 ==  bw
styincb

.elif @0 ==  cw
styincc

.endif
.endmacro


; **********************************************************************************************

;					Load a 16-bit register with a literal

;					All 16-bit registers r0w to r30w, AW, BW, CW, XW, YW, ZW


.macro litw			; <16_bit_reg> := literal 
.if @0   ==  r0w  	
lit r0,low(@1)
lit r1,high(@1)

.elif @0 ==  r2w
lit r2,low(@1)
lit r3,high(@1)

.elif @0 ==  r4w
lit r4,low(@1)
lit r5,high(@1)

.elif @0 ==  r6w
lit r6,low(@1)
lit r7,high(@1)

.elif @0 ==  r8w
lit r8,low(@1)
lit r9,high(@1)

.elif @0 ==  r10w
lit r10,low(@1)
lit r11,high(@1)

.elif @0 ==  r12w
lit r12,low(@1)
lit r13,high(@1)

.elif @0 ==  r14w
lit r14,low(@1)
lit r15,high(@1)

.elif @0 ==  r16w
ldi r16,low(@1)		
ldi r17,high(@1)

.elif @0 ==  r18w
ldi r18,low(@1)		
ldi r19,high(@1)

.elif @0 ==  r20w
ldi r20,low(@1)		
ldi r21,high(@1)

.elif @0 ==  r22w
ldi r22,low(@1)		
ldi r23,high(@1)

.elif @0 ==  r24w
ldi r24,low(@1)		
ldi r25,high(@1)

.elif @0 ==  r26w
ldi r26,low(@1)		
ldi r27,high(@1)

.elif @0 ==  r28w
ldi r28,low(@1)		
ldi r29,high(@1)

.elif @0 ==  r30w
ldi r30,low(@1)		
ldi r31,high(@1)

.endif
.endmacro


; **********************************************************************************************

; 					Operations with 16-bit literals (registers AW, BW, CW, XW, YW, ZW)


.macro addlitw			; Add. <16_bit_reg> + literal 
.if @0   ==  aw  		
addlita @1

.elif @0 ==  bw
addlitb @1

.elif @0 ==  cw
addlitc @1

.elif @0 ==  xw
addlitx @1

.elif @0 ==  yw
addlity @1

.elif @0 ==  zw
addlitz @1

.endif
.endmacro


.macro sublitw			; Subtract. <16_bit_reg> - literal 
.if @0   ==  aw  		
sublita @1

.elif @0 ==  bw
sublitb @1

.elif @0 ==  cw
sublitc @1

.elif @0 ==  xw
sublitx @1

.elif @0 ==  yw
sublity @1

.elif @0 ==  zw
sublitz @1

.endif
.endmacro


.macro complitw			; Compare by subtracting <16_bit_reg> + literal 
.if @0   ==  aw  		
complita @1

.elif @0 ==  bw
complitb @1

.elif @0 ==  cw
complitc @1

.elif @0 ==  xw
complitx @1

.elif @0 ==  yw
complity @1

.elif @0 ==  zw
complitz @1

.endif
.endmacro


.macro andlitw			; AND. <16_bit_reg> AND literal (bitwise) 
.if @0   ==  aw  		
andlita @1

.elif @0 ==  bw
andlitb @1

.elif @0 ==  cw
andlitc @1

.elif @0 ==  xw
andlitx @1

.elif @0 ==  yw
andlity @1

.elif @0 ==  zw
andlitz @1

.endif
.endmacro


.macro orlitw			; OR. <16_bit_reg> OR literal (bitwise) 
.if @0   ==  aw  		
orlita @1

.elif @0 ==  bw
orlitb @1

.elif @0 ==  cw
orlitc @1

.elif @0 ==  xw
orlitx @1

.elif @0 ==  yw
orlity @1

.elif @0 ==  zw
orlitz @1

.endif
.endmacro


.macro xorlitw			; XOR. <16_bit_reg> XOR literal (bitwise) 
.if @0   ==  aw  		
xorlita @1

.elif @0 ==  bw
xorlitb @1

.elif @0 ==  cw
xorlitc @1

.elif @0 ==  xw
xorlitx @1

.elif @0 ==  yw
xorlity @1

.elif @0 ==  zw
xorlitz @1

.endif
.endmacro



.macro zerow			; Test whether tne content of the 16_bit_reg is zero 
.if @0   ==  aw  		
zeroa @1

.elif @0 ==  bw
zerob @1

.elif @0 ==  cw
zeroc @1

.elif @0 ==  xw
zerox @1

.elif @0 ==  yw
zeroy @1

.elif @0 ==  zw
zeroz @1

.endif
.endmacro



